"""Python Module to contain the functions for performing resampling.

Resamplers all subclass ``pydpf.Module``
"""
import torch
from torch import Tensor
from typing import Tuple, Any
from .utils import batched_select
from .base import Module
from . import optimal_transport
from math import sqrt, log



class MultinomialResampler(Module):
    """Multinomial resampler.

        Resamples particles from the multinomial distribution. With

        Parameters
        ----------
        generator: torch.Generator
            The generator to track the random state of the resampling process.
    """
    def __init__(self, generator:torch.Generator):


        super().__init__()
        self.cache = {}
        self.generator = generator
        self._need_weight_output = True

    def forward(self, state:Tensor, weight:Tensor, **data):
        """Run the multinomial resampler."""
        with torch.no_grad():
            sampled_indices = torch.multinomial(torch.exp(weight), weight.size(1), replacement=True, generator=self.generator).detach()
            self.cache['used_weight'] = weight
            self.cache['sampled_indices'] = sampled_indices
        if self._need_weight_output:
            return batched_select(state, sampled_indices), torch.full(weight.size(), -log(weight.size(1)), device=weight.device, dtype=weight.dtype)
        return batched_select(state, sampled_indices)

class SystematicResampler(Module):
    """Systematic resampler.

        Under systematic resampling, the expected number of times a given particle is resampled is the same as for multinomial resampling. But it
        inter-correlates all the particles within a sample so it is difficult to provide the same theoretical guarantees on the asymptotic
        behaviour of filters that use systematic resampling compared to multinomial resampling. However, the stability offered by systematic
        resampling often results in better performance in practice.

        Parameters
        ----------
        generator: torch.Generator
            The generator to track the random state of the resampling process.

        Notes
        -----
        Equivalent to the Systematic resampler described in [1]_. Systematic resampling was first proposed in [2]_.
        Systematic resampling introduces strong dependence between particles and their index. Should the forward kernel be dependent on the
        particle index then the particles should be shuffled after resampling.

        References
        ----------
        .. [1] Chopin N, Papaspiliopoulos O (2020). An Introduction to Sequential Monte Carlo, chapter Importance Resampling, pp. 105–127. Springer.
        .. [2] Carpenter J, Clifford P, Fearnhead P (1999). “Improved particle filter for nonlinear problems.” In IEE Proc. Radar, Sonar and Navi., volume 146.
    """
    def __init__(self, generator:torch.Generator):
        super().__init__()
        self.cache = {}
        self.generator = generator
        self._need_weight_output = True

    def forward(self, state:Tensor, weight:Tensor, **data):
        """Run the systematic resampler."""
        with torch.no_grad():
            offset = torch.rand((weight.size(0),), device=state.device, generator=self.generator)
            cum_probs = torch.cumsum(torch.exp(weight), dim=1)
            # No index can be above 1. and the last index must be exactly 1.
            # Fix this in case of numerical errors
            cum_probs = torch.where(cum_probs > 1., 1., cum_probs)
            cum_probs[:, -1] = 1.
            resampling_points = torch.arange(weight.size(1), device=state.device) + offset.unsqueeze(1)
            sampled_indices = torch.searchsorted(cum_probs * weight.size(1), resampling_points)
            self.cache['used_weight'] = weight
            self.cache['sampled_indices'] = sampled_indices
        if self._need_weight_output:
            return batched_select(state, sampled_indices), torch.full(weight.size(), -log(weight.size(1)), device=weight.device, dtype=weight.dtype)
        return batched_select(state, sampled_indices)


class SoftResampler(Module):
    """Soft resampler.

    Module for perfoming soft-resampling, (P. Karkus, D. Hsu and W. S. Lee 'Particle Filter Networks with Application to
    Visual Localization' 2018).

    Soft resampling allows gradients to be passed through resampling by inducing importance weights. This is done by instead drawing the
    resampled particle from an alternative distribution and re-weighting the samples. The chosen alternative distribution is a mixture of
    the target with probability a; and a uniform distribution over the particles, with probability 1-a.

    The ``softness`` parameter, can be thought of as trading off between unbiased gradients (``softness`` = 0) and efficient resampling (``softness`` = 1). With
    ``softness`` > 0, the resampled index depends (randomly) on the previous weights. The contribution to the gradient from this dependence is ignored.

    Parameters
    ----------
    softness:  float
        The trade-off parameter between unbiased gradients (``softness`` = 0) and efficient resampling (``softness`` = 1).
    base_resampler: Module
        The base resampler to use.
    device: torch.device.
        The device that filtering is performed on

    Notes
    -----
    Proposed in [1]_. Like many of the resamplers in PyDPF this resampler acts on top of another resampler, generally this should be either ``MultinomialResampler``, ``SystematicResampler`` or ``OptimalTransportResampler``.
    Stacking other resamplers should be done with great care and in the order: -- Resamplers that modify gradient computation above Resamplers that modify the weights above Resamplers that modify the distribution
    for given weights) --. But it will nearly always be safer to define a new resampler with the desired behaviour than to stack exotic resamplers.

    References
    ----------
    .. [1] Karkus P, Hsu D, Lee WS (2018). “Particle filter networks with application to visual localization.” In Proc. Conf. Robot Learn., pp. 169–178. PMLR, Zurich, CH.
    """

    def __new__(cls, softness:float, base_resampler:Module, device:torch.device):
        if softness == 1:
            return base_resampler
        else:
            return super().__new__(cls)

    def __init__(self, softness:float, base_resampler:Module, device:torch.device):
        super().__init__()
        self.cache = {}
        self.softness = softness
        if softness < 0 or softness >= 1:
            raise ValueError(f'Softness {softness} is out of range, must be in [0,1)')
        self.log_softness = torch.log(torch.tensor(softness, device = device))
        self.neg_log_softness = torch.log(torch.tensor(1-softness ,device = device))
        self.resampler = base_resampler
        self._need_weight_output = True

    def forward(self, state:Tensor, weight:Tensor, **data):
        """Run the soft-resampler"""
        log_n = torch.log(torch.tensor(weight.size(1), device=state.device))
        soft_weight = torch.logaddexp(weight + self.log_softness, self.neg_log_softness - log_n)
        state, output_weights = self.resampler(state, soft_weight)
        self.cache = self.resampler.cache
        if self._need_weight_output:
            return state, batched_select(weight, self.cache['sampled_indices']) - batched_select(self.cache['used_weight'], self.cache['sampled_indices']) - log_n
        return state


class StopGradientResampler(Module):
    '''Stop-gradient resampling

    Stop-gradient resampling uses the REINFORCE or score-based Monte-Carlo gradient technique applied at each time-step.
    For numerical stability our implementation attaches the gradients to the log-space weights, rather than the linear-space particles.

    Parameters
    ----------
    generator: torch.Generator
        The generator to track the random state of the resampling process.

    Notes
    -----
    REINFORCE is unbiased after a single application but not so when applied in series as we do here when ``time_extent`` > 1.
    Proposed in [1]_. Like many of the resamplers in PyDPF this resampler acts on top of another resampler, generally this should be either ``MultinomialResampler``, ``SystematicResampler`` or ``OptimalTransportResampler``.
    Stacking other resamplers should be done with great care and in the order: -- Resamplers that modify gradient computation above Resamplers that modify the weights above Resamplers that modify the distribution
    for given weights) --. But it will nearly always be safer to define a new resampler with the desired behaviour than to stack exotic resamplers.

    References
    ----------
    .. [1] Scibior A, Wood F (2021). “Differentiable particle filtering without modifying the forward pass.” arXiv:2106.10314
    '''
    def __init__(self, base_resampler:Module):
        super().__init__()
        self.cache = {}
        self.base_resampler = base_resampler
        self._need_weight_output = True

    def forward(self, state:Tensor, weight:Tensor, **data):
        """Run the stop-gradient resampler"""

        if self._need_weight_output and torch.is_grad_enabled():
            self.base_resampler._need_weight_output = False
            state = self.base_resampler(state, weight)
            resampled_weight = batched_select(self.base_resampler.cache['used_weight'], self.base_resampler.cache['sampled_indices'])
            self.cache = self.base_resampler.cache
            self.cache['used_weight'] = self.cache['used_weight'].detach()
            return state, resampled_weight - resampled_weight.detach()

        if self._need_weight_output:
            self.base_resampler._need_weight_output = True
            state, new_weight = self.base_resampler(state, weight)
            self.cache = self.base_resampler.cache
            return state, new_weight

        self.base_resampler._need_weight_output = False
        state = self.base_resampler(state, weight)
        self.cache = self.base_resampler.cache
        return state

class OptimalTransportResampler(Module):
    r"""Optimal transport based resampling


        Optimal transport resampling produces a differentiable deterministic transport map from the proposal distribution to the posterior.
        This is achieved by finding the solution to an entropy regularised Kantorovich optimal transport problem between the two empirical
        distributions. The particles are transformed by the resulting optimal map to obtain a new unweighted approximation of the posterior.


        Parameters
        ----------
        regularisation: float
            The minimum strength of the entropy regularisation, in our implementation regularisation automatically chosen per sample and
             exponentially decayed to this value.
        decay_rate: float
            The factor by which to decrease the entropy regularisation per Sinkhorn loop.
        min_update_size: float
            The size of update to the transport potentials below which iteration should stop.
        max_iterations: int
            The maximum number iterations of the Sinkhorn loop, before stopping. Regardless of convergence.
        transport_gradient_clip: float
            The maximum per-element gradient of the transport matrix that should be passed. Higher valued gradients will be clipped to this value.

        Notes
        -----
        Our implementation is closely based on the original code of Thornton and Corenflos, the following details being taken from theirs:
        We exponentially decay the regularisation strength over the Sinkhorn iterations.
        We chose the initial strength of the regularisation parameter to be equal to maximum value minus the minimum value in the particle-state 2D array of the particle positions after each dimension is normalised to standard deviation 1.
        For numerical stability we cap the magnitude of the contribution to the gradient due to the transport matrix.

        Optimal transport resampling places particles in new positions on :math:`\mathbb{R}^n`, so it cannot directly be applied when some component of
        the state space is discrete/categorical.

        Optimal transport resampling results in biased (but asymptotically consistent) estimates of all non-affine functions of the latent state.
        Including the likelihood. The authors of the proposing paper investigate this effect and find it sufficiently small to ignore. See their
        paper [1]_ for details.
        """

    def __init__(self, regularisation: float, decay_rate: float, min_update_size: float, max_iterations: int, transport_gradient_clip: float):
        super().__init__()
        self.cache = {}
        self.regularisation = regularisation
        self.decay_rate = decay_rate
        self.min_update_size = min_update_size
        self.max_iterations = max_iterations
        self._need_weight_output = True

        class OTGradientWrapper(torch.autograd.Function):
            """
            Optimal transport gradient can suffer from numerical instability.
            Clip the gradient of the loss wrt the transport matrix to some user specified value.
            This is done in Corenflos and Thornton's original implementation.
            """

            @staticmethod
            def forward(ctx: Any, transport_matrix: Tensor):
                return transport_matrix

            @staticmethod
            def backward(ctx: Any, dtransport) -> Any:
                return torch.clip(dtransport, -transport_gradient_clip, transport_gradient_clip)

        self.gradient_wrapper = OTGradientWrapper

    @staticmethod
    def diameter(x: Tensor):
        """
        Calculates the diameter of the data.
        The diameter is defined as the maximum of the standard deviation across a sample across data dimensions.

        Parameters
        ----------
        x: Tensor
            Input tensor.

        Returns
        -------
        diameter: Tensor
            The diameter of the data per batch.
        """
        diameter_x = torch.amax(x.std(dim=1, unbiased=False), dim=-1, keepdim=True)
        return torch.where(torch.eq(diameter_x, 0.), 1., diameter_x)

    @staticmethod
    def extent(x: Tensor):
        return torch.amax(x, dim=(1,2)) - torch.amin(x, dim=(1,2))

    @staticmethod
    def get_sinkhorn_inputs_OT(Nk, log_weight: Tensor, x_t: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Get the inputs to the Sinkhorn algorithm as used for OT resampling

        Parameters
        ----------
        log_weights: (B,N) Tensor
            The particle weights

        N: int
            Number of particles

        x_t: (B,N,D) Tensor
            The particle state

        Returns
        -------
        log_uniform_weights: (B,N) Tensor
            A tensor of log(1/N)

        cost_matrix: (B, N, N) Tensor
            The auto-distance matrix of scaled_x_t under the 2-Norm.

        scale_x: (B, N, D) Tensor
            The amount the particles where scaled by in calculating the cost matrix.
        """
        log_uniform_weight = torch.log(torch.ones((log_weight.size(0), Nk), device=log_weight.device) / Nk)
        centred_x_t = x_t - torch.mean(x_t, dim=1, keepdim=True).detach()
        scale_x = OptimalTransportResampler.diameter(x_t).detach() * sqrt(x_t.size(-1))
        scaled_x_t = centred_x_t / scale_x.unsqueeze(2)
        cost_matrix = torch.cdist(scaled_x_t, scaled_x_t, 2) ** 2
        extent = OptimalTransportResampler.extent(scaled_x_t)
        return log_uniform_weight, cost_matrix, extent


    def forward(self, state: Tensor, weight: Tensor, **data):
        """Run the optimal transport resampler."""
        N = state.size(1)
        log_b, cost, extent = self.get_sinkhorn_inputs_OT(N, weight, state)
        f, g, epsilon_used = optimal_transport.sinkhorn_loop(weight, log_b, cost, self.regularisation, self.min_update_size, self.max_iterations, extent.reshape(-1, 1, 1), self.decay_rate)
        transport = optimal_transport.get_transport_from_potentials(weight, log_b, cost, f, g, self.regularisation)
        transport = self.gradient_wrapper.apply(transport)
        self.cache['used_weight'] = weight
        if self._need_weight_output:
            return optimal_transport.apply_transport(state, transport, N), torch.full(weight.size(), -log(weight.size(1)), device=weight.device, dtype=weight.dtype)
        return optimal_transport.apply_transport(state, transport, N)

class KernelResampler(Module):


    def __init__(self, kernel):
        """
            Returns a function for performing differentiable kernel resampling (Younis and Sudderth 'Differentiable and Stable Long-Range Tracking of Multiple Posterior Modes' 2024).

            Notes
            -----
            Unlike the majority of implemented resampling schemes this function returns a Module object so that learnable parameters may be properly registered.
            In this case the final returned field is always a tensor containing a single zero.

            Parameters
            ----------
            kernel: Distribution
                The kernel to convolve over the particles to form the KDE sampling distribution.
            generator: torch.Generator
                The generator to track the random state of the resampling process.

            Returns
            -------
            kernel_resampler: Resampler:
                A Module whose forward method implements kernel resampling.
        """
        super().__init__()
        self.mixture = kernel
        self.cache = {}
        self._need_weight_output = True

    def forward(self, state: Tensor, weight: Tensor, **data):
        """Run the kernel transport resampler."""
        with torch.no_grad():
            new_state = self.mixture.sample(state, weight, sample_size=(state.size(1),))
            self.cache = self.mixture.resampler.cache
        # Save computation if gradient is not required
        if torch.is_grad_enabled() and self._need_weight_output:
            density = self.mixture.log_density(new_state, state, weight)
            new_weight = density - density.detach()
            return new_state, new_weight

        if self._need_weight_output:
            return new_state, torch.full(weight.size(), -log(weight.size(1)), device=weight.device, dtype=weight.dtype)

        return new_state


"""Module for loading and saving files, not part of the public API"""
import polars as pl
from pathlib import Path
from joblib import Parallel, delayed
import numpy as np
from typing import Tuple, Union


def _load_directory_csv(
    directory: Path,
    *,
    processes=-1,
) -> Tuple[pl.DataFrame, pl.DataFrame]:
    """Load and concatenate CSV files from a directory into a single DataFrame.

    This function reads all CSV files in the specified directory, adds a
    'series_id' column to each DataFrame based on the file name, and then
    concatenates them into a single DataFrame. The resulting DataFrame is
    sorted by 'series_id', while retaining the order of other columns.

    Parameters
    ----------
    directory (Path): The path to the directory containing the CSV files.
    processes (int, optional): The number of processes to use for parallel
        reading. Defaults to -1, which means using all available processors.

    Returns
    -------
    pl.DataFrame: A DataFrame containing data from all CSV files
    in the directory, with an added 'series_id' column.
    """

    def read_helper(file: Path):
        file_data = pl.read_csv(file)
        file_data = file_data.with_columns(pl.lit(file.stem).alias("series_id"))
        return file_data

    read_output = list(
        Parallel(n_jobs=processes)(
            delayed(read_helper)(file)
            for file in directory.iterdir()
            if file.suffix == ".csv" and not file.suffix == "series_metadata.csv"
        )
    )
    return pl.concat(read_output).sort("series_id", maintain_order=True)

def _load_series_metadata_csv(file: Path):
    series_metadata = pl.read_csv(file)
    try:
        series_metadata.get_column('series_id')
        series_metadata = series_metadata.sort('series_id', ascending=True)
    except:
        series_metadata.with_row_index('series_id', 1)
    return series_metadata


def _load_file_csv(
    file: Path,
) -> pl.DataFrame:
    return pl.read_csv(file)


def _extract_tensor(
    data: pl.DataFrame, prefix: str, series_id_column: str, *, prefix_is_complete=False
) -> np.ndarray:
    """
    Extracts tensor data from a given DataFrame by selecting columns that match a specified prefix and grouping by a series ID column.

    Parameters
    ----------
    data (pl.DataFrame): The input DataFrame containing the data.
    prefix (str): The prefix of the columns we wish to extract.
    series_id_column (str): The column name used to group the data.

    Returns
    -------
    np.ndarray: An AxBxC tensor, with the (a,b,c)th element being the cth element of the bth timestep of the ath series.
        A is the number of series (number of unique elements of the series_id_column),
        B is the number of timesteps in each series, must be the same for all series (no padding is implemented yet),
        C is the number of elements in each timestep (number of columns prefixed by prefix).
    """
    if not prefix_is_complete:
        data = data.select([pl.col(f"^{prefix}.*$"), series_id_column])
    else:
        data = data.select([pl.col(prefix), series_id_column])
    data_gb = data.group_by(series_id_column, maintain_order=True)
    tensor_list = []
    # key_list = []
    for gb_key, series_group_data in data_gb:
        tensor = series_group_data.select(pl.exclude(series_id_column)).to_numpy()
        tensor_list.append(tensor)
        # key_list.append(gb_key[0])
    # key_array = np.array(key_list)
    tensor_array = np.stack(tensor_list)
    return tensor_array


def load_data_csv(
    data_path: Union[Path, str],
    *,
    series_metadata_path : Union[Path, str, None] = None,
    series_id_column="series_id",
    state_prefix=None,
    observation_prefix="observation",
    time_column=None,
    control_prefix=None,
):
    if isinstance(data_path, str):
        data_path = Path(data_path)
    if isinstance(series_metadata_path, str):
        series_metadata_path = Path(series_metadata_path)
    series_metadata = None
    if data_path.is_dir():
        data = _load_directory_csv(data_path)
        if series_metadata_path is not None:
            series_metadata = _load_series_metadata_csv(series_metadata_path)
        else:
            try:
                series_metadata = _load_series_metadata_csv(data_path / "series_metadata.csv")
            except FileNotFoundError:
                pass
    elif data_path.is_file():
        data = _load_file_csv(data_path)
        if series_metadata_path is not None:
            series_metadata = _load_series_metadata_csv(series_metadata_path)
    else:
        raise ValueError("Invalid data path")

    tensor_dict = {}


    if state_prefix is not None:
        tensor_dict["state"] = _extract_tensor(data, state_prefix, series_id_column)

    tensor_dict["observation"] = _extract_tensor(
        data, observation_prefix, series_id_column
    )

    if control_prefix is not None:
        tensor_dict["control"] = _extract_tensor(data, control_prefix, series_id_column)

    if time_column is not None:
        tensor_dict["time"] = _extract_tensor(data, time_column, series_id_column, prefix_is_complete=True)

    current_index_end = 0
    output_dict = {}
    if series_metadata is not None:
        output_dict["series_metadata"] = series_metadata.to_numpy()
    else:
        output_dict["series_metadata"] = None
    output_dict["indices"] = {}
    tensor_list = []
    for key, tensor in tensor_dict.items():
        tensor_list.append(tensor)

        new_index_end = current_index_end + tensor.shape[-1]

        output_dict["indices"][key] = range(current_index_end, new_index_end)

        current_index_end = new_index_end

    output_dict["tensor"] = np.concatenate(tensor_list, axis=-1)

    return output_dict



from .base import Distribution
from torch import Tensor
from typing import Union
import torch
from ..utils import batched_select
from ..resampling import MultinomialResampler
from ..base import Module


class CompoundDistribution(Distribution):
    r"""A wrapper that concatenates several distributions into a single distribution object.
        Interdependence between the distributions is not permitted.

        Parameters
        ----------
        distributions: list[Distribution|Module, ...]
            An iterable of the distributions that form the components of the compound.
        generator : Unused parameter

        Notes
        -----
        If a distribution in the list distributions is not a Distribution subclass (i.e. it is a custom implementation) then it must have the following attributes::

        dim : the dimension of the distribution.

        sample() : method that takes the parameter `sample_size` and returns a tensor of that size with an extra final dimension of size dim.

        log_density() : method that takes the parameter `sample` and returns a tensor of the same size without the final dimension, assumed to be of size dim.
    """
    conditional = False

    def __init__(self, distributions: list[Distribution|Module], generator):
        super().__init__(generator)
        self.dists = distributions
        self.dim = 0
        #register submodules
        self.dists = torch.nn.ModuleList(distributions)
        for dist in self.dists:
            if not isinstance(dist, Distribution):
                if not hasattr(dist, "sample"):
                    raise AttributeError('Custom distributions must have an "sample" method.')
                if not hasattr(dist, "log_density"):
                    raise AttributeError('Custom distributions must have an "log_density" method.')
                if not hasattr(dist, "dim"):
                    raise AttributeError('Custom distributions must have an "dim" attribute.')
            else:
                if type(dist).conditional:
                    raise TypeError(f'None of the component distributions may be conditional, detected {type(dist)} which is.')
                if self.device != dist.device:
                    raise ValueError(f'All component distributions must have all parameters on the same device, found {self.device} and {dist.device}.')
            self.dim += dist.dim


    def sample(self, sample_size: tuple[int, ...]|None) -> Tensor:
        """Sample a Compound distribution.
        The sample is the concatenation of samples from the components distributions along the last axis.

        Parameters
        ----------
        sample_size : tuple[int, ...]|None
            The size of the sample to draw. If None then a single sample is drawn and no sample dimension is used.

        Returns
        -------
        sample: Tensor
            The resulting sample.
        """
        samples = []
        for dist in self.dists:
            samples.append(dist.sample(sample_size))
        return torch.cat(samples, dim=-1)

    def log_density(self, sample: Tensor) -> Tensor:
        """Evaluate the log density of a sample.

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.

        Returns
        -------
        sample log_density: Tensor
            The log density of each datum in the sample.

        """
        output = None
        dim_count = 0
        for dist in self.dists:
            if output is None:
                output = dist.log_density(sample[..., dim_count:dim_count+dist.dim])
            else:
                output += dist.log_density(sample[..., dim_count:dim_count+dist.dim])
            dim_count += dist.dim
        return output


def sample_only_multinomial(state: Tensor, weights: Tensor, generator) -> Tensor:
    with torch.no_grad():
        sampled_indices = torch.multinomial(torch.exp(weights), weights.size(1), replacement=True, generator=generator).detach()
    return batched_select(state, sampled_indices)


class KernelMixture(Distribution):
    """Create a kernel density mixture.
    The parameter kernel is an unconditional distribution which will be convolved over the kernel density mixture.
    The resultant distribution is conditional on the locations and weights of the kernels.

    Parameters
    ----------
    kernel: list[tuple[str,int], ...]|Distribution
        The kernel to convolve over the particles to form the KDE sampling distribution.
    generator : Union[torch.Generator, None]
        The generator to control the rng when sampling kernels from the mixture.

    Notes
    -----
    If the kernel is not a Distribution subclass (i.e. it is a custom implementation) then it must have the following attributes:

    dim : the dimension of the distribution.

    sample() : method that takes the parameter `sample_size` and returns a tensor of that size with an extra final dimension of size dim.

    log_density() : method that takes the parameter `sample` and returns a tensor of the same size without the final dimension, assumed to be of size dim.
    """
    conditional = True

    def __init__(self, kernel: Distribution|Module, generator: Union[torch.Generator, None]):
        super().__init__(generator)
        self.resampler = MultinomialResampler(generator=generator)
        if not isinstance(kernel, Distribution):
            if not hasattr(kernel, "sample"):
                raise AttributeError('Custom distributions must have an "sample" method.')
            if not hasattr(kernel, "log_density"):
                raise AttributeError('Custom distributions must have an "log_density" method.')
            if not hasattr(kernel, "dim"):
                raise AttributeError('Custom distributions must have an "dim" attribute.')
        self.kernel = kernel
        self.dim = self.kernel.dim
        if isinstance(kernel, Distribution) and type(self.kernel).conditional:
            raise ValueError(f'The kernel distribution cannot be conditional, detected {type(self.kernel)} which is.')

    def _check_conditions(self, loc: Tensor, weight: Tensor) -> None:
        if loc.device != self.device:
            raise ValueError(f'loc must be on the same device as the distribution, found {loc.device} and {self.device}.')
        if weight.device != self.device:
            raise ValueError(f'weight must be on the same device as the distribution, found {weight.device} and {self.device}.')
        if loc.size(-1) != self.dim:
            raise ValueError(f'It is not permitted for the kernel to have a different dimension as the space it is convolved over, found {loc.size(-1)} and {self.dim}.')
        if weight.size(-1) != loc.size(-2):
            raise ValueError(f'Differing number of kernels locations{loc.size(-2)} and weights {weight.size(-1)}.')


    def sample(self,  loc: Tensor, weight: Tensor, sample_size: tuple[int,...]|None) -> Tensor:
        """Sample a KDE mixture

        Parameters
        ----------
        loc : Tensor
            Locations of the Kernels
        weight  : Tensor
            Weights of the Kernels
        sample_size : tuple[int,...]|None
            The size of the sample to draw. If None then a single sample is drawn per batch dimension and no sample dimension is used.

        Returns
        -------
        Sample: Tensor
            A sample from the KDE mixture.
        """
        #Multinomial resampling is sampling from a KDE with a dirac kernel.
        self._check_conditions(loc, weight)
        try:
            sampled_locs, _ = self.resampler(loc, weight, generator=self.generator)
        except Exception as e:
            raise RuntimeError(f'Failed to sample kernels with error: \n {e} \n This is likely due to a mismatch in batch dimensions.')
        batch_size = self.get_batch_size(sampled_locs.size(), 2)
        if sample_size is None:
            sample = self.kernel.sample(sample_size=batch_size)
        else:
            sample = self.kernel.sample(sample_size=(*batch_size, *sample_size))
        return sampled_locs + sample

    def log_density(self, sample: Tensor, loc: Tensor, weight: Tensor) -> Tensor:
        """Evaluate the log density of a sample.

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.
        loc : Tensor
            Locations of the Kernels
        weight  : Tensor
            Weights of the Kernels.

        Returns
        -------
        Sample: Tensor
            The log density of each datum in the sample.
        """
        try:
            self._check_conditions(loc, weight)
            densities = self.kernel.log_density(sample.unsqueeze(-2) - self._unsqueeze_to_size(loc, sample.unsqueeze(-2), 2))
            return torch.logsumexp(densities + weight.unsqueeze(-2), dim=-1)
        except RuntimeError as e:
            raise RuntimeError(f'Failed to apply condition with error: \n {e} \n This is likely to a mismatch in batch dimensions.')


import torch
from torch import Tensor
from typing import Callable
from .base import Module
from .distributions import KernelMixture

class FilteringMean(Module):
    """Get an estimate of the filtering mean of a function of the latent state.

    Parameters
    ----------
    function: Callable[[Tensor], Tensor]. Default identity function.
        The function of the latent state to estimate.

    Notes
    -----
    function can be (explicitly) a function of the state only.
    """
    def __init__(self,function: Callable[[Tensor], Tensor] = lambda x: x):
        super().__init__()
        self.function = function

    def forward(self, *, state: Tensor, weight: Tensor, **data) -> Tensor:
        """Get the filtering mean for a given time step."""
        return torch.einsum('ij..., ij -> i... ', self.function(state), torch.exp(weight))

class MSE_Loss(Module):
    """Get the per-timestep mean squared error of a function of the latent state compared to ground truth over a batch of filters.

    Parameters
    ----------
    function: Callable[[Tensor], Tensor]. Default identity function.
        The function of the latent state to estimate.
    """
    def __init__(self,function: Callable[[Tensor], Tensor] = lambda x: x):

        super().__init__()
        self.mean = FilteringMean(function)

    def forward(self, *, state: Tensor, weight: Tensor, ground_truth, **data):
        """Get the MSE loss for a time-step"""
        filter_mean = self.mean(state = state, weight = weight)
        return torch.mean(torch.sum((ground_truth - filter_mean) ** 2, dim=-1))

class LogLikelihoodFactors(Module):

    def __init__(self):
        """Get the log observation likelihood factor for each time step. Such that an estimate of the log likelihood over the trajectory may be given by a sum of these factors."""
        super().__init__()

    def forward(self, likelihood, **kwargs) -> Tensor:
        """Get the log likelihood factor for a time step."""
        return likelihood

class ElBO_Loss(Module):
    """Get the factors of the ELBO loss per-timestep for a batch of filters. The complete ELBO loss may be given by a sum of these factors. This is the negative of the ELBO.

    Notes
    -----
    As an average of the log-likelihood, the ELBO estimates a Jensens' inequality lower bound to the mean log of the likelihood.
    The SMC ELBO is not exactly analogous to the variation auto-encoder ELBO, see [1]_ for more information.

    References
    ----------
    .. [1] Naesseth C, Linderman S, Ranganath R, Blei D (2018). “Variational sequential monte carlo.” In Proc. Int. Conf. Art. Int. and Stat. (AISTATS), pp. 968–977. PMLR, Lanzarote, Canary Islands.
    """
    def __init__(self):
        super().__init__()

    def forward(self, likelihood, **kwargs):
        """Get the negative ELBO factor for a time step."""
        return -torch.mean(likelihood)


class PredictiveMean(Module):
    """Predict the state n steps ahead.

    Parameters
    ----------
    prediction_kernel: Module
        A function to importance sample from the predictive distribution n-steps ahead. Typically, this will entail be applying the bootstrap proposal n-times.
    lag: int
        How many steps ahead the prediction is being made.
    function: Callable[[Tensor], Tensor]
        The function of the latent state to estimate.
    """
    def __init__(self, prediction_kernel: Callable, lag: int, function: Callable[[Tensor], Tensor] = lambda x: x):
        super().__init__()
        self.prediction_kernel = prediction_kernel
        self.lag = lag
        self.function = function

    def forward(self, *, state: Tensor, norm_weights: Tensor, data, time, **kwargs):
        prediction, new_weights = self.prediction_kernel(state, norm_weights, data[time:time+self.lag].squeeze(), time)
        return torch.einsum('ij...,ij...->i...', self.function(prediction), torch.exp(new_weights))


class NegLogDataLikelihood_Loss(Module):
    """Get the negative log data likelihood per-timestep under a kernel density estimator.
    This function applies a kernel density estimator over the particles and calculates the log likelihood of the ground truth given the KDE.

    Parameters
    ----------
    kernel: KernelMixture
        The kernel density estimator.
    """
    def __init__(self, kernel: KernelMixture):

        super().__init__()
        self.KDE = kernel


    def forward(self, *, state: Tensor, weight: Tensor, ground_truth, **kwargs):
        """Get the negative log data likelihood factor under the a KDE and given a time-step"""
        return -self.KDE.log_density(ground_truth, state, weight)

class State(Module):
    """Module that returns the particle states when passed as an `aggregation_function`."""

    def __init__(self):
      super().__init__()

    def forward(self, *, state: Tensor, **data):
        """Return the particle states"""
        return state


class Weight(Module):
    """Module that returns the particle weights when passed as an `aggregation_function`."""
    def __init__(self):
        super().__init__()

    def forward(self, *, weight: Tensor, **data):
        """Return the particle weights"""
        return weight


from .base import *
from .Gaussian import MultivariateGaussian, LinearGaussian, ConditionalGaussian, StandardGaussian
from .kde import KernelMixture, CompoundDistribution
from .vonMises import VonMises


"""This Module implements methods for modifying the gradient of the particles. This is most often done for stability."""

from typing import Tuple, Any
import torch
from torch import Tensor

def ClipByElement(clip_threshold: float) -> torch.autograd.Function:
    """Clips the per-element gradient of the loss due to the state/weights to some constant value at each time-step.

        Parameters
        ----------
        clip_threshold: float
        The threshold above which to clip.
    """
    class _ClipByElement(torch.autograd.Function):
        @staticmethod
        def forward(ctx: Any, state:Tensor, weight:Tensor, time_data: dict) -> Tuple[Tensor, Tensor]:
            return state, weight

        @staticmethod
        def backward(ctx: Any, dstate:Tensor, dweight:Any):
            return torch.clip(dstate, clip_threshold, clip_threshold), torch.clip(dweight, -clip_threshold, clip_threshold), None

    return _ClipByElement

def ClipByNorm(clip_threshold: float) -> torch.autograd.Function:
    """Clips the gradient of the particles and their weights to some constant value.

        Parameters
        ----------

        clip_threshold: float
        The threshold above which to clip.
    """
    class _ClipByNorm(torch.autograd.Function):
        @staticmethod
        def forward(ctx: Any, state: Tensor, weight: Tensor, time_data: dict) -> Tuple[Tensor, Tensor]:
            return state, weight

        @staticmethod
        def backward(ctx: Any, dstate: Tensor, dweight: Tensor):
            dstate_norm = torch.linalg.vector_norm(dstate)
            dweight_norm = torch.linalg.norm(dweight)
            if dstate_norm > clip_threshold:
                dstate = dstate/dstate_norm
            if dweight_norm > clip_threshold:
                dweight = dweight/dweight_norm
            return dstate, dweight , None
    return _ClipByNorm


def ClipByParticle(clip_threshold: float) -> torch.autograd.Function:
    """Clips the norm of the gradient assigned per-particle.

    Parameters
    ----------

    clip_threshold: float
    The threshold above which to clip.
    """
    class _ClipByParticle(torch.autograd.Function):
        @staticmethod
        def forward(ctx: Any, state: Tensor, weight: Tensor, time_data: dict) -> Tuple[Tensor, Tensor]:
            ctx.save_for_backward(state, weight)
            return state, weight

        @staticmethod
        def backward(ctx: Any, dstate: Tensor, dweight: Tensor):
            state, weight = ctx.saved_tensors
            exp_weights = torch.exp(weight).unsqueeze(-1)
            dparticle = dstate/exp_weights + dweight.unsqueeze(-1)/(exp_weights*state)
            mag_dparticle = torch.linalg.vector_norm(dparticle, -1, keepdim=True)/2
            zero_mask = (exp_weights > 1e-7)
            too_big_mask = torch.logical_and(mag_dparticle > clip_threshold, zero_mask)
            dstate = torch.where(zero_mask, torch.where(too_big_mask, dstate/mag_dparticle, dstate), 0.)
            dweight = torch.where(zero_mask.squeeze(), torch.where(too_big_mask.squeeze(), dweight/mag_dparticle.squeeze(), dweight), 0.)
            return dstate, dweight, None
    return _ClipByParticle

def TBPTT(propagation_length: int):
    """Truncates the computation graph every ``propagation_length`` time-steps.

    Parameters
    ----------
    propagation_length: int
        The amount of time-steps after which to halt the gradient flow.
    """

    #Truncating can be done more efficiently on the forward pass, so create a class that looks like an autograd function
    #to the filtering script but only acts on the forward pass.
    class _TBPTT:
        @staticmethod
        def apply(state: Tensor, weight: Tensor, time_data: dict):
            if time_data['t'] % propagation_length == 0:
                return state.detach(), weight.detach()
            return state, weight

    return _TBPTT




"""Utility module for handling state-space data"""
import torch
import pandas as pd
from torch.utils.data import Dataset
from typing import Union, Callable, Tuple
import os
import numpy as np
from torch import Tensor
from math import ceil
from joblib import Parallel, delayed
from .deserialisation import load_data_csv
from .model_based_api import FilteringModel
from pathlib import Path
from itertools import chain

class StateSpaceDataset(Dataset):
    """Dataset class for state-observation data.

        Latent state of the system stored in the state Tensor.

        Dimensions are Discrete Time - Batch - Data

        When used as called from a dataloader you must use the custom collate function
        Data will always be returned in the order 'state' - 'observation' - 'time' - 'control' - 'metadata'

        At the moment only functionality to load entire data set into RAM/VRAM is provided.
        Lazy loading is a planned feature.

        Parameters
        ----------
        data_path: Union[Path,str].
            The path of the data file or folder.
        series_id_column: str. Default "series_id"
             The heading of the series_id column in the csv files.
        state_prefix: str|None. Default None.
            The prefix of heading of the state columns in the csv files.
        observation_prefix: str. Default "observation".
            The prefix of heading of the observation columns in the csv files.
        time_column: str|None. Default None.
            The heading of the time column in the csv files.
        control_prefix: str|None. Default None.
            The prefix of heading of the control columns in the csv files.
        device: torch.device. Default torch.device('cpu').

        Notes
        -----
        We provide methods to load data from files, obeying a certain format,
        into a map-style ``torch.utils.data.Dataset`` object and therefore be accessed easily from a
        ``torch.utils.data.DataLoader``. We allow one of two data storage formats, either storing
        the entire dataset in a single .csv file, or storing each trajectory in separate files {1.csv,
        2.csv, ..., T.csv} in a dedicated directory. The .csv files are formed of headed columns
        there must be at least one observation column, with state, time, and control columns
        being optional. As all the data categories, apart from time, are vector valued there can be
        multiple columns for each category. For the single-file format there must be additionally a
        series_id column that will be used to index each trajectory, for the multiple file format the
        series_id is encoded in the file name.
        The data category series_metadata exists to store exogenous variables that the trajectories
        might depend on, but are constant over a trajectory. These are to be stored in a separate
        .csv indexed by a series_id column.
        Given a file in the required format, loading a dataset is simple: initialise this class
        with the data’s path, the column labels and the device to store data retrieved by the data
        loader. When initialising the data loader, it is crucial that the argument collate_fn is set to
        ``dataset.collate`` where dataset is the dataset passed to the data loader. PyTorch’s default
        collate function will not return the data in a format that obeys PyDPF conventions. When
        looping over the data loader, data is returned as tuple in the ordering state - observation -
        time - control - series_metadata with only the field that exist being returned.


        See test_trajectory.csv at https://github.com/John-JoB/pydpf/tree/main/jss_examples/Stochastic%20Volatility for an example.

        .. Note:: When initialising a ``torch.utils.data.DataLoader`` the argument collate_fn must be set to ``dataset.collate`` where ``dataset`` is the instance of this class passed to the data loader.
    """

    @property
    def state(self):
        if 'state' in self.data_order:
            return self.data['tensor'][:, :, self.data['indices']['state']].permute(1, 0, 2).contiguous()
        raise AttributeError('No state data available')

    @property
    def observation(self):
        if 'observation' in self.data_order:
            return self.data['tensor'][:, :, self.data['indices']['observation']].permute(1, 0, 2).contiguous()
        raise AttributeError('No state data available')

    @property
    def time(self):
        if 'time' in self.data_order:
            return self.data['tensor'][:, :, self.data['indices']['time']].squeeze(-1).permute(1, 0).contiguous()
        raise AttributeError('No time data available')

    @property
    def control(self):
        if 'control' in self.data_order:
            return self.data['tensor'][:, :, self.data['indices']['control']].permute(1, 0, 2).contiguous()
        raise AttributeError('No control data available')

    @property
    def metadata(self):
        if self.metadata:
            return self.data['metadata']
        raise AttributeError('No metadata data available')

    def __init__(self,
                 data_path: Union[Path,str],
                 *,
                 series_id_column: str = "series_id",
                 state_prefix: str|None = None,
                 observation_prefix: str ="observation",
                 time_column: str|None =None,
                 control_prefix: str|None =None,
                 device: torch.device = torch.device('cpu')
             ):
        self.device = device
        self.data = load_data_csv(data_path, series_id_column = series_id_column, state_prefix = state_prefix, observation_prefix = observation_prefix, time_column = time_column, control_prefix = control_prefix)
        self.data['tensor'] = torch.from_numpy(self.data['tensor']).to(device=self.device, dtype=torch.float32)
        self.data_order = []
        if state_prefix is not None:
            self.data_order.append('state')
        self.data_order.append('observation')
        if time_column is not None:
            self.data_order.append('time')
        if control_prefix is not None:
            self.data_order.append('control')
        self.metadata_exists = False
        try:
            self.series_metadata = torch.from_numpy(self.data['metadata']).to(device=self.device, dtype=torch.float32)
            self.metadata_exists = True
        except KeyError:
            pass

    @property
    def observation_dimension(self):
        return self.observation.shape[-1]

    @property
    def state_dimension(self):
        return self.state.shape[-1]

    @property
    def control_dimension(self):
        return self.control.shape[-1]

    def __len__(self):
        return self.data['tensor'].size(0)

    def __getitem__(self, idx):
        if self.metadata_exists:
            return self.data['tensor'][idx], self.data['metadata'][idx]
        return self.data['tensor'][idx]

    def collate(self, batch) -> Tuple[torch.Tensor, ...]:
        """Pass to the ``collate_fn`` parameter of any ``torch.utils.data.DataLoader`` object that uses this dataset."""
        #By default, the batch is the first dimension.
        #Pass this function to collate_fn when defining a dataloader to make it the second.
        #collated_batch = torch.utils.data.default_collate(batch)
        if self.metadata_exists:
            batch = tuple(zip(*batch))
            collated_data = torch.stack(batch[0], dim=0).transpose(0, 1)
            collated_metadata = torch.stack(batch[1], dim=0)
            return (*(collated_data[:, :, self.data['indices'][data_category]].squeeze(-1).contiguous() if data_category == "time" else collated_data[:, :, self.data['indices'][data_category]].contiguous() for data_category in self.data_order),
                    collated_metadata)
        else:
            collated_batch = torch.stack(batch, dim=0).transpose(0, 1)
            return *(collated_batch[:, :, self.data['indices'][data_category]].squeeze(-1).contiguous() if data_category == "time" else collated_batch[:, :, self.data['indices'][data_category]].contiguous() for data_category in self.data_order),

    def normalise_dims(self, normalised_series:str = 'observation', scale_dims: str = 'all', individual_timesteps: bool = False, dims: Union[Tuple[int], None] = None) -> Tuple[torch.Tensor, torch.Tensor]:
        """Normalise the data to have mean zero and standard deviation one.

        This function normalises the data inplace and returns the offset and scale. Such that the original data can be reclaimed by original_data = normalised_data * scale + offset.

        This function can be applied to either the state or observations, this is controlled by the parameter normalise_state.

        There are various methods to control the scaling, determined by the value of scale_dims:
            - 'all': scale each dimension independently, such that every dimension have standard deviation 1.
            - 'max': scale each dimension by the same factor, such that the maximum of the standard deviations is 1.
            - 'min': scale each dimension by the same factor, such that the minimum of the standard deviations is 1.
            - 'norm': scale each dimension by the same factor, such that the standard deviation of the vector norm of the data is 1.

        The parameter individual_timesteps controls whether to apply the same normalisation across time-steps, or to calculate a separate mean and standard deviation per time-step.

        The normalisation doesn't have to be across all data dimensions, one can specify a tuple of dimensions to include to the parameter dims. Or set dims=None to use all dimensions.

        Parameters
        ----------
        normalise_state: bool
            When True, normalise the state. When False, normalise the observations.
        scale_dims: str
            The method to scale over dimensions. See above for options and details.
        individual_timesteps: bool, default=True
            When true, the scaling and offset is calculated per-time-step, when false the scaling and offset are set to be the same for each time-step (in most cases this should be True).
        dims: Tuple[int] or None, default=None
            The dimensions to normalise.

        Returns
        -------
        offset: torch.Tensor
            The per-element offset.
        scaling: torch.Tensor
            The per-element scaling.
        """
        with torch.no_grad():
            if not scale_dims in ['all', 'max', 'min', 'norm']:
                raise ValueError('scale_dims must be one of "all", "max", "min" or "norm"')

            data = self.data['tensor'][:, :, self.data['indices'][normalised_series]].clone()
            data_size = data.size(-1)
            data = data.transpose(0, -1)
            if dims is None:
                mask = [True for _ in range(data_size)]
            else:
                mask = [False for _ in range(data_size)]
                for d in dims:
                    if d < 0:
                        d = data_size + d
                    if d >= data_size or d < 0:
                        raise IndexError('Dimension out of bounds')
                    mask[d] = True

            if individual_timesteps:
                reduction_dims = (2,)
            else:
                reduction_dims = (1, 2)

            masked_data = data[mask]

            means = torch.mean(masked_data, dim=reduction_dims, keepdim=True)
            if scale_dims == 'all':
                std = torch.std(masked_data, dim=reduction_dims, keepdim=True)
            if scale_dims == 'max':
                std = torch.amax(torch.std(masked_data, dim=reduction_dims, keepdim=True), dim=0, keepdim=True)
            if scale_dims == 'min':
                std = torch.amin(torch.std(masked_data, dim=reduction_dims, keepdim=True), dim=0, keepdim=True)
            if scale_dims == 'norm':
                std = torch.std(torch.linalg.vector_norm(masked_data, dim=0, keepdim=True), dim=reduction_dims, keepdim=True)

            means = means.expand(masked_data.size())
            std = std.expand(masked_data.size())
            data[mask] = (masked_data - means) / std
            self.apply(lambda **data_dict : data.transpose(0,1), modified_series=normalised_series)
            return means.transpose(0, -1).transpose(0,1).contiguous(), std.transpose(0, -1).transpose(0,1).contiguous()

    def apply(self, f, modified_series:str = 'observation'):
        """Apply a function across all trajectories

        Takes a function f that takes a ``**dictionary`` of data categories, e.g. ``f = lambda: time, state, **kwargs = time * state`` for a function that returns the state multiplied by the time.
        And replaces the series given by ``modified_series`` with the output of f for every trajectory in a dataset.

        Parameters
        ----------
        f: function
            function to be applied across all trajectories
        modified_series: str. Default 'observation'
            The series to replace with the output of f

        """
        with torch.no_grad():
            true_order = ['state', 'observation', 'time', 'control', 'metadata']
            if not modified_series in true_order:
                raise ValueError('modified_series must be one of "state", "observation", "control", "time", or "metadata"')

            partitioned_data = {data_category: self.data['tensor'][:, :, self.data['indices'][data_category]].transpose(0,1).contiguous() for data_category in self.data_order}
            if self.metadata_exists:
                partitioned_data['metadata'] = self.data['metadata']
            new_series = f(**partitioned_data).transpose(0,1)
            if modified_series in self.data_order:
                inverse_index = [i for i in range(self.data['tensor'].size(-1)) if (i not in self.data['indices'][modified_series])]
                new_data = self.data['tensor'][:, :, inverse_index]
                if new_series.dim() == 2:
                    new_series = new_series.unsqueeze(-1)
                start_index = new_data.size(-1)
                self.data['tensor'] = torch.cat((new_data, new_series), dim=-1)
                for series in self.data_order:
                    if self.data['indices'][series][0] > self.data['indices'][modified_series][0]:
                        self.data['indices'][series] = range(self.data['indices'][series][0]-new_series.size(-1), self.data['indices'][series][-1] + 1 - new_series.size(-1))
                self.data['indices'][modified_series] = range(start_index, self.data['tensor'].size(-1))
                return
            self.data_order = [series for series in true_order if (series in self.data_order or series == modified_series)]
            start_index = self.data['tensor'].size(-1)
            self.data['tensor'] = torch.cat((self.data['tensor'], new_series), dim=-1)
            self.data['indices'][modified_series] = range(start_index, self.data['tensor'].size(-1))






def _get_time_data(data: dict, t: int) -> dict:
    time_dict = {k:v[t] for k, v in data.items() if k != 'series_metadata'}
    try:
        time_dict['series_metadata'] = data['series_metadata']
    except KeyError:
        pass
    return time_dict


def _format_to_save(state, observation, control, time):
    data_list = [state.cpu().numpy(), observation.cpu().numpy()]
    columns_list = [[f'state_{i + 1}' for i in range(state.size(-1))], [f'observation_{i + 1}' for i in range(observation.size(-1))]]
    if control is not None:
        data_list.append(control.cpu().numpy())
        columns_list.append([f'control_{i + 1}' for i in range(state.size(-1))])
    if time is not None:
        data_list.append(time.unsqueeze(-1).cpu().numpy())
        columns_list.append(['time'])
    return np.concatenate(data_list, axis=-1), list(chain.from_iterable(columns_list))

def _save_directory_csv(path:Path, start_index, state, observation, control, time, n_processes = -1):

    data, columns_list = _format_to_save(state, observation, control, time)
    def write_help(series_id):
        df = pd.DataFrame(data[series_id - start_index])
        df.columns = columns_list
        df.to_csv(path / f'trajectory_{series_id + 1}.csv' ,  index=False)
    Parallel(n_jobs=n_processes)(delayed(write_help)(series_id)
                                 for series_id in range(start_index, start_index + state.size(0))
                                 )

def _save_file_csv(path:Path, state, observation, control, time, n_processes = -1):
    data, columns_list = _format_to_save(state, observation, control, time)
    def make_traj_frame(series_id):
        df = pd.DataFrame(data[series_id])
        df.columns = columns_list

        df['series_id'] = series_id + 1
        return df
    #df_list = list(Parallel(n_jobs=n_processes)(delayed(make_traj_frame)(series_id)
                                               # for series_id in range(len(data))
                                                #))
    df_list = [make_traj_frame(series_id) for series_id in range(len(data))]
    total_df = pd.concat(df_list, axis=0)
    total_df.to_csv(path, index=False)


def simulate_and_save(data_path: Union[Path, str],
                    SSM: FilteringModel,
                    *,
                    time_extent: int,
                    n_trajectories: int,
                    batch_size: int,
                    device: Union[str, torch.device] = torch.device('cpu'),
                    control: Tensor = None,
                    time:Tensor = None,
                    n_processes = -1,
                    bypass_ask = False):

    r"""Simulate data from a state-space model or and save as csv in our standard format or a collection of csvs, one per trajectory.

        Parameters
        ----------
        data_path: Union[Path, str]
            The path at which to save the generated data. Can be either a ``.csv`` or directory.
        SSM: FilteringModel
            The state-space model to simulate from.
        time_extent: int
            The amount of time steps to simulate per-trajectory excluding ``t=0``. Taking ``time_extent = T`` generates data at time-steps :math:`t \in [0,\dots,T]`
        n_trajectories: int
            The number of the trajectories to simulate.
        batch_size: int
            The number of trajectories to simulate as a batch, using GPU parallelism if available.
        device: Union[str, torch.device]
            The device to generate the trajectories on. Default CPU.
        control: Tensor|None, Default: None
            The control actions.
        time: Tensor|None, Default: None
            The time each timestep should occur at.
        n_processes: int, Default = -1
            The number of cpu processes to use to save the data in parallel. When equal to -1 all cores are used.
        bypass_ask: bool, Default = False
            If False then this function will ask for confirmation before it overwrites an existing file or directory. If bypass_ask is True then it proceed without asking for confirmation.

        Notes
        -----
        If ``data_path`` ends in ".csv" then all trajectories will be saved in a single csv file at that path. If it is a directory then the trajectories will be saved in separate csvs in that directory.

        ``SSM`` must have the following components ``prior_model``, ``dynamic_model`` and ``observation_model``. All must have a ``.sample`` method defined. Despite not generating multiple samples per-trajectory,
        as we do during filtering all ``.sample`` methods should act the as if there were. I.e. the particle dimension should exist, but will always be of size 1.

        ``control`` should be a ``time_extent`` X ``n_trajectories`` X ``inherent_dim`` tensor if specified. ``time`` should be a ``time_extent`` X ``n_trajectories`` if specified. This matches usual ``PyDPF`` convention.

        ``n_processes`` only specifies the number of processes used for saving the data and not generating it. Generating the data is done using standard CUDA parallelism if on the GPU and no parallelism on the CPU. This
        is efficient on the GPU and avoids the pitfalls of multiprocessing with PyTorch.
    """

    prior = lambda _batch_size, **_data_dict:  torch.squeeze(SSM.prior_model.sample(_batch_size, 1, **_data_dict), 1)
    observation_model = lambda _state, **_data_dict: SSM.observation_model.sample(state=_state, **_data_dict)
    Markov_kernel = lambda _prev_state, **_data_dict: SSM.dynamic_model.sample(prev_state=_prev_state, **_data_dict)
    if isinstance(data_path, str):
        data_path = Path(data_path)
    if data_path.suffix == '.csv':
        state_list = []
        observation_list = []
        if data_path.is_file():
            if not bypass_ask:
                print(f'Warning - file already exists at {data_path}, continuing could overwrite its data')
                response = input('Continue? (y/n) ')
                if response != 'Y' and response != 'y':
                    print('Halting')
                    return
            os.remove(data_path)
    else:
        if data_path.is_dir() and not bypass_ask:
            print(f'Warning - folder already exists at {data_path}, continuing could overwrite its data')
            response = input('Continue? (y/n) ')
            if response != 'Y' and response != 'y':
                print('Halting')
                return
        else:
            os.mkdir(data_path)

    data_dict = {}

    n_batches = ceil(n_trajectories / batch_size)

    with torch.inference_mode():
        for batch in range(n_batches):
            print(f'Generating batch {batch + 1}/{n_batches}', end = '\r')
            if batch == (n_trajectories // batch_size):
                if control is not None:
                    batch_control = control[batch * batch_size:]
                    data_dict['control'] = batch_control[:, 0]
                if time is not None:
                    batch_time = time[batch * batch_size:]
                    data_dict['time'] = batch_time[: 0]
                print()
                temp = prior(n_trajectories - batch*batch_size, **data_dict)
            else:
                if control is not None:
                    batch_control = control[batch * batch_size : (batch + 1) * batch_size]
                    data_dict['control'] = batch_control[:, 0]
                if time is not None:
                    batch_time = time[batch * batch_size : (batch + 1) * batch_size]
                    data_dict['time'] = batch_time[: 0]
                temp = prior(batch_size, **data_dict)
            state = torch.empty(size=(temp.size(0), time_extent+1, temp.size(1)), dtype=torch.float32, device=device)
            state[:, 0] = temp
            temp = observation_model(state[:, 0], **data_dict)
            observation = torch.empty(size=(temp.size(0), time_extent+1, temp.size(1)), device=device)
            observation[:, 0] = temp
            for t in range(time_extent):
                if control is not None:
                    data_dict['control'] = batch_control[:, t]
                if time is not None:
                    data_dict['time'] = batch_time[:, t]
                state[:, t+1] = Markov_kernel(state[:, t], **data_dict)
                observation[:, t+1] = observation_model(state[:, t+1], **data_dict)
            if data_path.suffix == '.csv':
                state_list.append(state)
                observation_list.append(observation)
            else:
                _save_directory_csv(data_path, batch_size*batch, state, observation, control, time, n_processes)
        if data_path.suffix == '.csv':
            state = torch.cat(state_list, dim=0)
            observation = torch.cat(observation_list, dim=0)
            _save_file_csv(data_path, state, observation, control, time)
    print('Done                  \n')











"""Python module for the core filtering algorithms."""

import torch
from torch import Tensor
from .utils import normalise
from .base import Module
from .resampling import SystematicResampler, SoftResampler, OptimalTransportResampler, StopGradientResampler, KernelResampler, MultinomialResampler
from .distributions import KernelMixture
from .model_based_api import FilteringModel
from .base import DivergenceError
from warnings import warn
from .conditional_resampling import ConditionalResampler
from copy import copy
from typing import Callable



class SIS(Module):
    """Module that represents a sequential importance sampling (SIS) algorithm. A SIS algorthm is fully specified by its importance sampling
        procedures, the user should supply a proposal kernel that may depend on the time-step; and a special case for time 0.

    Parameters
    ----------
    initial_proposal: Module | None
        A callable object that takes the number of particles and the data/observations at time-step zero and returns an importance sample
        of the posterior, i.e. particle position and log weights. Also returns the observation likelihood (if applicable).
    proposal: Module | None
        A callable object that implements the proposal kernel. Takes the state and log weights at the previous time step,
        the discreet time index i.e. how many iterations the filter has run for; and the data/observations at the current time-step.
        And returns an importance sample of the posterior at the current time step, i.e. particle position and log weights.
        Also returns the observation factor likelihood.

    Notes
    -----
    SMC filters can, in general, be described as special cases of sequential importance sampling (SIS).
    We provide this generic SIS class that can be extended for a given use case, or used by directly supplying the relevant functions.
    SIS iteratively importance samples a Markov-Chain.
    An SIS algorithm is defined by supplying an initial distribution and a Markov kernel.

    This implementation is more general than the standard SIS algorithm. There is no independence requirements for the samples within a
    batch. This means that the particles can be drawn from an arbitrary joint distribution on depended on the data and the particles at
    the previous time-step. Both the usual particle filter [1]_ and interacting multiple model particle filter [2]_ are special
    cases of this algorithm.

    References
    ----------
    .. [1] Chopin N, Papaspiliopoulos O (2020). An Introduction to Sequential Monte Carlo, chapter
       Particle Filtering, pp. 129–165. Springer.

    .. [2] Boers Y, Driessen J (2003). “Interacting multiple model particle filter.” IEE Proc. Radar,
       Sonar Nav., 150, 344–349. ISSN 1350-2395.
    """
    def __init__(self, *, initial_proposal: Module | None = None, proposal: Module | None = None):
        super().__init__()
        if initial_proposal is not None:
            self._register_functions(initial_proposal, proposal)


    def _register_functions(self, initial_proposal: Module, proposal: Module):
        self.initial_proposal = initial_proposal
        self.proposal = proposal
        self.aggregation_function = None

    @staticmethod
    def _get_time_data(t: int, **data) -> dict:
        time_dict = {k:v[t] for k, v in data.items() if k != 'series_metadata' and k != 'state' and v is not None}
        time_dict['t'] = t
        if data['time'] is not None and t>0:
            time_dict['prev_time'] = data['time'][t-1]
        if data['series_metadata'] is not None:
            time_dict['series_metadata'] = data['series_metadata']
        return time_dict

    class _EmptyGradReg:
        @staticmethod
        def apply(state, weight, time_data):
            return state, weight

    def forward(self,
                n_particles: int,
                time_extent: int,
                aggregation_function: dict[str, Module] | Module,
                observation: Tensor,
                *,
                gradient_regulariser: torch.autograd.Function | None = None,
                ground_truth: Tensor | None = None,
                control: Tensor | None = None,
                time: Tensor | None = None,
                series_metadata: Tensor | None = None) -> dict[str, Tensor] | Tensor :
        """Run a forward pass of the SIS filter.

        Parameters
        ----------
        n_particles: int
            The number of particles to draw per filter.
        time_extent: int
            The maximum time-step to run to, including time 0, the filter will draw {time_extent + 1} importance sample populations.
        aggregation_function: dict[str, Module] or Module
            A module that's forward function processes the filtering outputs (the particle locations, the normalised log weights,
            the log sum of the unnormalised weights, the data, the time-step) into an output per time-step.
            Or a string indexed dictionary of such items.
        observation: Tensor
            The observations of the hidden variable system.
        gradient_regulariser: torch.autograd.Function or None. Default None.
            A autograd function to apply to the particles at the start of every time-step. It should leave the forward pass unchanged but may modify the gradients during the backward pass
            the intended usage is to regularise the gradient in some way. Optional.
        ground_truth: Tensor or None. Default None.
            The ground truth latent state. Should only be pass to the aggregation function and not used in the proposal functions. Optional.
        control: Tensor or None. Default None.
            The control actions. Optional.
        time: Tensor or None. Default None.
            The continuous time each time-step occurs at. Optional.
        series_metadata: Tensor or None. Default None.
            The series_metadata. Optional.

        Returns
        -------
        output: Tensor or dict[str, Tensor]
            The output of the filter, formed from stacking the output of aggregation_function for every time-step.
            Or if aggregation_function is a dictionary, a dictionary of these output Tensors one for each aggregation function.

        Notes
        -----
        To save memory during inference runs we allow the user to pass a function that takes a population
        of particles and processes this into an output for each time-step. For example, if the goal was the filtering mean then it would be
        wasteful to store the full population of the particles for every time-step. The memory savings are most impactful during inference as
        pytorch retains many intermediate tensors for gradient computation otherwise.
        """
        #Register module should the aggregation function have learnable parameters
        if isinstance(aggregation_function,dict):
            self.aggregation_function = torch.nn.ModuleDict(aggregation_function)
            output_dict = True
        else:
            self.aggregation_function = aggregation_function
            output_dict = False

        if gradient_regulariser is None or not (self.training and torch.is_grad_enabled()):
            gradient_regulariser = self._EmptyGradReg

        gt_exists = False
        if ground_truth is not None:
            gt_exists = True
        time_data = self._get_time_data(0, observation = observation, control = control, time = time, series_metadata = series_metadata)
        state, weight, likelihood = self.initial_proposal(n_particles = n_particles, **time_data)
        likelihood = likelihood - torch.log(torch.tensor(n_particles, dtype=torch.float32, device=observation.device))

        if output_dict:
            output = {}
            for name, function in aggregation_function.items():
                if gt_exists:
                    temp = function(state=state, weight=weight, likelihood=likelihood, ground_truth=ground_truth[0], **time_data)
                else:
                    temp = function(state=state, weight=weight, likelihood=likelihood, **time_data)
                output[name] = torch.empty((time_extent+1, *temp.size()), device = observation.device, dtype=torch.float32)
                output[name][0] = temp
        else:
            if gt_exists:
                temp = aggregation_function(state = state, weight = weight, likelihood = likelihood, ground_truth = ground_truth[0], **time_data)
            else:
                temp = aggregation_function(state = state, weight = weight, likelihood = likelihood, **time_data)
            output = torch.empty((time_extent+1, *temp.size()), device = observation.device, dtype=torch.float32)
            output[0] = temp
        for t in range(1, time_extent+1):
            try:
                time_data = self._get_time_data(t, observation = observation, control = control, time = time, series_metadata = series_metadata)
                prev_state, prev_weight = gradient_regulariser.apply(state, weight, time_data)
                state, weight, likelihood = self.proposal(prev_state = prev_state, prev_weight = prev_weight, **time_data)
                if output_dict:
                    for name, function in aggregation_function.items():
                        if gt_exists:
                            output[name][t] = function(state=state, weight=weight, likelihood=likelihood, ground_truth=ground_truth[t], **time_data)
                        else:
                            output[name][t] = function(state=state, weight=weight, likelihood=likelihood, **time_data)
                else:
                    if gt_exists:
                        output[t] = aggregation_function(state=state, weight=weight, likelihood=likelihood, ground_truth = ground_truth[t], **time_data)
                    else:
                        output[t] = aggregation_function(state=state, weight=weight, likelihood=likelihood, **time_data)
            except DivergenceError as e:
                warn(f'Detected divergence at time-step {t} with message:\n    {e} \nStopping iteration early.')
                if isinstance(output, dict):
                    return {name: value[:t-1] for name, value in output.items()}
                else:
                    return output[:t-1]
        return output


class ParticleFilter(SIS):
    """The standard SIR particle filter


    Parameters
    ----------
    resampler: Module
        The resampling algorithm to use. Takes the state and log weights at the previous time-step and returns the state and log weights
        after resampling.
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    use_REINFORCE_for_proposal: bool
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process. Applying REINFORCE to only some components of the
        state space is not permitted with this API, such a use case would require a custom SIS process.

    Notes
    -----
        The standard particle filter is a special case of the SIS algorithm. We construct the particle filtering proposal by first
        resampling particles from their population, then applying a proposal kernel restricted such that the particles depend only on the
        population at the previous time-step through the particle at the same index [1]_. This Class forms the base for the majority of
        filtering algorithms packaged with PyDPF.

    References
    ----------
    .. [1] Chopin N, Papaspiliopoulos O (2020). An Introduction to Sequential Monte Carlo, chapter Particle Filtering, pp. 129–165. Springer.
    """
    def __init__(self, resampler: Module = None, SSM: FilteringModel = None, *, use_REINFORCE_for_proposal:bool=False, use_REINFORCE_for_initial_proposal:bool=False) -> None:
        self.REINFORCE_i = use_REINFORCE_for_initial_proposal
        self.REINFORCE = use_REINFORCE_for_proposal

        super().__init__()
        if resampler is not None:
            self._register_functions(resampler=resampler, SSM=SSM)

    @staticmethod
    def _make_PF_prior(SSM, REINFORCE_i:bool):
        if SSM.has_initial_proposal:
            if REINFORCE_i:
                def prior(n_particles, observation, **data):
                    with torch.no_grad():
                        state = SSM.initial_proposal_model.sample(batch_size = observation.size(0), n_particles = n_particles, observation=observation, **data)
                        prop_density = SSM.initial_proposal_model.log_density(state = state, observation = observation, **data)
                    weight = (SSM.observation_model.score(state = state, observation = observation, **data)
                              - prop_density
                              + SSM.prior_model.log_density(state = state, **data))
                    return state, weight
            else:
                def prior(n_particles, observation, **data):
                    state = SSM.initial_proposal_model.sample(batch_size = observation.size(0), n_particles = n_particles, observation=observation, **data)
                    weight = (SSM.observation_model.score(state = state, observation = observation, **data)
                              - SSM.initial_proposal_model.log_density(state = state, observation = observation, **data)
                              + SSM.prior_model.log_density(state = state, **data))
                    return state, weight
        else:
            if REINFORCE_i:
                def prior(n_particles, observation, **data):
                    with torch.no_grad():
                        state = SSM.prior_model.sample(batch_size = observation.size(0), n_particles = n_particles, **data)
                    density = SSM.prior_model.log_density(state = state, **data)
                    weight = SSM.observation_model.score(state = state, observation = observation, **data) + density - density.detach()
                    return state, weight
            else:
                def prior(n_particles, observation, **data):
                    state = SSM.prior_model.sample(batch_size = observation.size(0), n_particles = n_particles, **data)
                    weight = SSM.observation_model.score(state = state, observation = observation, **data)
                    return state, weight

        return prior

    @staticmethod
    def _make_PF_proposal(SSM, REINFORCE:bool):
        if SSM.has_proposal:
            if REINFORCE:
                def prop(prev_state, prev_weight, observation, **data):
                    with torch.no_grad():
                        new_state = SSM.proposal_model.sample(prev_state = prev_state, observation=observation, **data)
                        prop_density = SSM.proposal_model.log_density(state = new_state, prev_state = prev_state, observation = observation, **data)
                    new_weight = (prev_weight + SSM.observation_model.score(state = new_state, observation = observation, **data)
                                  - prop_density
                                  + SSM.dynamic_model.log_density(state = new_state, prev_state = prev_state, **data))
                    return new_state, new_weight
            else:
                def prop(prev_state, prev_weight, observation, **data):
                    new_state = SSM.proposal_model.sample(prev_state = prev_state, observation=observation, **data)
                    new_weight = (prev_weight + SSM.observation_model.score(state = new_state, observation = observation, **data)
                                  - SSM.proposal_model.log_density(state = new_state, prev_state = prev_state, observation = observation, **data)
                                  + SSM.dynamic_model.log_density(state = new_state, prev_state = prev_state, **data))
                    return new_state, new_weight
        else:
            if REINFORCE:
                def prop(prev_state, prev_weight, observation, **data):
                    with torch.no_grad():
                        new_state = SSM.dynamic_model.sample(prev_state = prev_state, **data)
                    density = SSM.dynamic_model.log_density(state=new_state, prev_state=prev_state, **data)
                    new_weight = prev_weight + SSM.observation_model.score(state=new_state, observation = observation, **data) + density - density.detach()
                    return new_state, new_weight
            else:
                def prop(prev_state, prev_weight, observation, **data):
                    new_state = SSM.dynamic_model.sample(prev_state = prev_state, **data)
                    new_weight = prev_weight + SSM.observation_model.score(state=new_state, observation = observation, **data)
                    return new_state, new_weight

        return prop


    def _register_functions(self, resampler: Module, SSM: FilteringModel):
        self.SSM = SSM
        self.resampler = resampler

        if self.REINFORCE:
            if not hasattr(SSM.dynamic_model, 'log_density'):
                raise AttributeError("The dynamic model must implement a 'log_density' method for REINFORCE.")
        if self.REINFORCE_i:
            if not hasattr(SSM.prior_model, 'log_density'):
                raise AttributeError("The prior model must implement a 'log_density' method for REINFORCE.")

        prior = ParticleFilter._make_PF_prior(SSM, self.REINFORCE_i)
        prop = ParticleFilter._make_PF_proposal(SSM, self.REINFORCE)

        def initial_sampler(n_particles: int, **data):
            state, weight = prior(n_particles=n_particles, **data)
            weight, likelihood = normalise(weight)
            return state, weight, likelihood

        def pf_sampler(prev_state, prev_weight, **data):
            resampled_x, resampled_w = self.resampler(prev_state, prev_weight, **data)
            state, weight = prop(prev_state=resampled_x, prev_weight=resampled_w, **data)
            try:
                weight, likelihood = normalise(weight)
            except ValueError:
                raise DivergenceError('Found batch where all weights are small.')
            return state, weight, likelihood
        super()._register_functions(initial_sampler, pf_sampler)


class MarginalParticleFilter(SIS):
    """The marginal particle filter

    Parameters
    ----------
    resampler: Module
        The resampling algorithm to use. Takes the state and log weights at the previous time-step and returns the state and log weights
        after resampling.
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.
    optimise_for_bootstrap: bool. Default True.
        If the system is bootstrap then the filter can be optimised due to not computing terms which are guaranteed to cancel. Set to True to take advantage of these optimisations.

    Notes
    -----
    The marginal particle filter is a special case of the SIS algorithm. The unlike standard particle filter the marginal particle filter considers resampling in its evaluation of the proposal distribution by
    accounting for the possibility that any new particle could have been derived from any particle at the previous time-step instead of only considering it's genealogical path. This is done by marginalising over the
    ancestral indices at each time-step, hence the name. [1]_

    The optimisations taken when ``optimise_for_bootstrap`` is True are only valid if the algorithm is truly bootstrap aside from the first time-step, i.e. the proposal and resampling both do not induce importance
    weights. However, it is permitted for ``SSM.initial_proposal_model`` to be non-bootstrap. If ``SSM.proposal_model`` is non-None then this is detected and the value of ``optimise_for_bootstrap`` is ignored.
    However, there is no efficient way to detect if the resampler modifies the weights so it is on the user to manually set ``optimise_for_bootstrap`` to False in this case, otherwise this filter will silently use
    an algorithm with undefined behaviour.

    .. warning:: Setting the parameter ``optimise_for_bootstrap`` to True can silently invoke undefined behaviour if the resampler is non-standard, see the Notes section for detail.


    References
    ----------
    .. [1] Klaas M, de Freitas N, Doucet A (2005). “Toward Practical N 2 Monte Carlo: the Marginal
       Particle Filter.” In Proc. Conf. Uncert. Art. Intell. (UAI), pp. 308–315. Arlington, Virginia.
    """


    def __init__(self,
                 resampler: Module|None = None,
                 SSM: FilteringModel = None,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False,
                 optimise_for_bootstrap: bool = True):
        super().__init__()
        self.REINFORCE_i = use_REINFORCE_for_initial_proposal
        self.REINFORCE = use_REINFORCE_for_proposal
        self.optimise = optimise_for_bootstrap and (not SSM.has_proposal)
        if self.optimise:
            self.PF = ParticleFilter(resampler, SSM)
        if resampler is not None:
            self._register_functions(resampler=resampler, SSM=SSM)


    def forward(self, *args, **kwargs):
        if self.optimise and not torch.is_grad_enabled():
            return self.PF.forward(*args, **kwargs)
        return super().forward(*args, **kwargs)


    @staticmethod
    def _make_MPF_proposal(SSM, resampler, REINFORCE, optimise) -> Callable:

        if SSM.has_proposal:
            if REINFORCE:
                def prop(prev_state, prev_weight, observation, **data):
                    resampled_state, resampled_weight = resampler(prev_state, prev_weight, **data)
                    used_weight = resampler.cache['used_weight']
                    with torch.no_grad():
                        state = SSM.proposal_model.sample(prev_state=resampled_state, observation=observation, **data).detach()
                    expanded_state = state.unsqueeze(2).expand(-1, -1, state.size(1), -1).flatten(1, 2)
                    expanded_prev_state = prev_state.unsqueeze(1).expand(-1, state.size(1), -1, -1).flatten(1, 2)
                    dynamic_log_density = SSM.dynamic_model.log_density(state=expanded_state, prev_state=expanded_prev_state, **data).reshape(state.size(0), state.size(1), state.size(1))
                    with torch.no_grad():
                        proposal_log_density = SSM.proposal_model.log_density(state=expanded_state, prev_state=expanded_prev_state, observation=observation, **data).reshape(state.size(0), state.size(1), state.size(1))
                    obs_score = SSM.observation_model.score(state=state, observation=observation, **data)
                    weight = (torch.logsumexp(prev_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                              - torch.logsumexp(used_weight.unsqueeze(1) + proposal_log_density, dim=-1)
                              + obs_score)
                    return state, weight, obs_score, dynamic_log_density, proposal_log_density, resampled_weight
            else:
                def prop(prev_state, prev_weight, observation, **data):
                    resampled_state, resampled_weight = resampler(prev_state, prev_weight, **data)
                    used_weight = resampler.cache['used_weight']
                    state = SSM.proposal_model.sample(prev_state=resampled_state, observation=observation, **data)
                    expanded_prev_state = prev_state.unsqueeze(1).expand(-1, state.size(1), -1, -1).flatten(1, 2)
                    expanded_state = state.unsqueeze(2).expand(-1, -1, state.size(1), -1).flatten(1, 2)
                    dynamic_log_density = SSM.dynamic_model.log_density(state=expanded_state, prev_state=expanded_prev_state, **data).reshape(state.size(0), state.size(1), state.size(1))
                    proposal_log_density = SSM.proposal_model.log_density(state=expanded_state, prev_state=expanded_prev_state, observation=observation, **data).reshape(state.size(0), state.size(1), state.size(1))
                    obs_score = SSM.observation_model.score(state=state, observation=observation, **data)
                    weight = (torch.logsumexp(prev_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                              - torch.logsumexp(used_weight.unsqueeze(1) + proposal_log_density, dim=-1)
                              + obs_score)
                    return state, weight, obs_score, dynamic_log_density, proposal_log_density, resampled_weight

        else:
            if REINFORCE:
                def prop(prev_state, prev_weight, observation, **data):
                    resampled_state, resampled_weight = resampler(prev_state, prev_weight, **data)
                    used_weight = resampler.cache['used_weight']
                    with torch.no_grad():
                        state = SSM.dynamic_model.sample(prev_state=resampled_state, **data)
                    expanded_prev_state = prev_state.unsqueeze(1).expand(-1, state.size(1), -1, -1).flatten(1, 2)
                    expanded_state = state.unsqueeze(2).expand(-1, -1, state.size(1), -1).flatten(1, 2)
                    dynamic_log_density = SSM.dynamic_model.log_density(state=expanded_state, prev_state=expanded_prev_state, **data).reshape(state.size(0), state.size(1), state.size(1))
                    obs_score = SSM.observation_model.score(state=state, observation=observation, **data)
                    detach_dynamic = dynamic_log_density.detach()
                    weight = (torch.logsumexp(prev_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                              - torch.logsumexp(used_weight.unsqueeze(1) + detach_dynamic, dim=-1)
                              + obs_score)
                    return state, weight, obs_score, dynamic_log_density, detach_dynamic, resampled_weight
            else:
                if optimise:
                    def prop(prev_state, prev_weight, observation, **data):
                        resampled_state, resampled_weight = resampler(prev_state, prev_weight, **data)
                        used_weight = resampler.cache['used_weight']
                        state = SSM.dynamic_model.sample(prev_state=resampled_state, **data)
                        expanded_prev_state = prev_state.unsqueeze(1).expand(-1, state.size(1), -1, -1).flatten(1, 2)
                        expanded_state = state.unsqueeze(2).expand(-1, -1, state.size(1), -1).flatten(1, 2)
                        obs_score = SSM.observation_model.score(state=state, observation=observation, **data)
                        # Detach to save computing redundant gradient terms.
                        with torch.no_grad():
                            dynamic_log_density = SSM.dynamic_model.log_density(state=expanded_state, prev_state=expanded_prev_state, **data).reshape(state.size(0), state.size(1), state.size(1))
                        weight = (torch.logsumexp(prev_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                                  - torch.logsumexp(used_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                                  + obs_score)
                        return state, weight, obs_score, 0, 0, resampled_weight
                else:
                    def prop(prev_state, prev_weight, observation, **data):
                        resampled_state, resampled_weight = resampler(prev_state, prev_weight, **data)
                        used_weight = resampler.cache['used_weight']
                        state = SSM.dynamic_model.sample(prev_state=resampled_state, **data)
                        expanded_prev_state = prev_state.unsqueeze(1).expand(-1, state.size(1), -1, -1).flatten(1, 2)
                        expanded_state = state.unsqueeze(2).expand(-1, -1, state.size(1), -1).flatten(1, 2)
                        dynamic_log_density = SSM.dynamic_model.log_density(state=expanded_state, prev_state=expanded_prev_state, **data).reshape(state.size(0), state.size(1), state.size(1))
                        obs_score = SSM.observation_model.score(state=state, observation=observation, **data)
                        weight = (torch.logsumexp(prev_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                                  - torch.logsumexp(used_weight.unsqueeze(1) + dynamic_log_density, dim=-1)
                                  + obs_score)
                        return state, weight, obs_score, 0, 0, resampled_weight

        return prop



    def _register_functions(self, resampler: Module, SSM: FilteringModel):
        self.resampler = resampler
        self.SSM = SSM

        prior = ParticleFilter._make_PF_prior(SSM, self.REINFORCE_i)
        prop = MarginalParticleFilter._make_MPF_proposal(SSM, resampler, self.REINFORCE_i, self.optimise)



        def initial_sampler(n_particles: int, **data):
            state, weight = prior(n_particles=n_particles, **data)
            weight, likelihood = normalise(weight)
            return state, weight, likelihood

        if isinstance(self.resampler, ConditionalResampler):
            def mpf_sampler(prev_state, prev_weight, observation, **data):
                state, weight, obs_score, dynamic_log_density, proposal_log_density, resampled_weight = prop(prev_state, prev_weight, observation, **data)
                weight = torch.where(self.resampler.cache['mask'], weight, resampled_weight + obs_score + torch.diag(dynamic_log_density) - torch.diag(proposal_log_density))
                try:
                    weight, likelihood = normalise(weight)
                except ValueError:
                    raise DivergenceError('Found batch where all weights are small.')

                return state, weight, likelihood
        else:
            def mpf_sampler(prev_state, prev_weight, observation, **data):
                state, weight, _, _, _, _ = prop(prev_state, prev_weight, observation, **data)
                try:
                    weight, likelihood = normalise(weight)
                except ValueError:
                    raise DivergenceError('Found batch where all weights are small.')
                return state, weight, likelihood

        super()._register_functions(initial_sampler, mpf_sampler)



class DPF(ParticleFilter):
    """Basic 'differentiable' particle filter.

    Parameters
    ----------
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    resampling_generator: torch.Generator
        Generator to track the resampling rng.
    multinomial: bool. Default False.
        If true then use multinomial resampling. Otherwise, use systematic resampling.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

        Notes
        -----
        This DPF does not use a differentiable resampling method, instead the gradients are detached at each time-step. See [1]_ for detail.

        References
        ----------
        .. [1] Jonschkowski R, Rastogi D, Brock O (2018). “Differentiable Particle Filters: End-to-End
           Learning with Algorithmic Priors.” In Proc. Robot.: Sci. Syst. Pittsburgh, PA, USA.
        """

    def __init__(self,
                 SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial:bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False
                 ) -> None:

        resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)
        temp = copy(self.proposal)
        self.proposal = lambda prev_state, prev_weight, **data: temp(prev_state.detach(), prev_weight.detach(), **data)

class SoftDPF(ParticleFilter):
    """Differentiable particle filter with soft-resampling. [1]_

    Parameters
    ----------
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    resampling_generator: torch.Generator
        Generator to track the resampling rng.
    multinomial: bool. Default False.
        If true then use multinomial resampling. Otherwise, use systematic resampling.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

    Notes
    -----
    Applies a non-target proposal to resampling, replace the resampling weights :math:`w^{i}_{t}` with:

    .. math:: (\text{softness}) w^{i}_{t} + \frac{(1-\text{softness})}{\text{n_particles}}

    With ``softness`` = 1 then this is equivelent to the straight-through estimator used in [2]_.

    References
    ----------
    .. [1] Karkus P, Hsu D, Lee WS (2018). “Particle filter networks with application to visual localization.” In Proc. Conf. Robot Learn., pp. 169–178. PMLR, Zurich, CH.

    .. [2] Naesseth C, Linderman S, Ranganath R, Blei D (2018). “Variational sequential monte carlo.” In Proc. Int. Conf. Art. Int. and Stat. (AISTATS), pp. 968–977. PMLR, Lanzarote, Canary Islands.
    """

    def __init__(self, SSM: FilteringModel = None,
                 softness: float = 0.7,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial: bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False,
                 ) -> None:

        base_resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        resampler = SoftResampler(softness, base_resampler, resampling_generator.device)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)

class OptimalTransportDPF(ParticleFilter):
    """Differentiable particle filter with optimal transport resampling.

    Parameters
    ----------
    SSM: FilteringModel
        A FilteringModel that represents the SSM (and optionally a proposal model). See the documentation of FilteringModel for more complete information.
    regularisation: float
        The maximum strength of the entropy regularisation, in our implementation the initial regularisation automatically chosen per sample and exponentially decreased to the given regularisation.
    decay_rate: float
        The factor by which to decrease the entropy regularisation per Sinkhorn loop.
    min_update_size: float
        The size of update to the transport potentials below which the algorithm is considered converged and iteration should stop.
    max_iterations: int
        The maximum number iterations of the Sinkhorn loop, before stopping. Regardless of convergence.
    transport_gradient_clip: float
        The maximum per-element gradient of loss due to the transport matrix that should be passed. Higher valued gradients will be clipped to this value.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

    Notes
    -----
    See [1]_. It is not recommended to set ``use_REINFORCE_for_proposal`` to True as it will lose the convergence properties of the reparameterised algorithm but retain the instability and costly runtime of OT filtering.
    Consider other algorithms if the proposal cannot be reparameterised.

    References
    ----------
    .. [1] Corenflos A, Thornton J, Deligiannidis G, Doucet A (2021). “Differentiable Particle Filtering via Entropy-Regularized Optimal Transport.” In Proc. Int. Conf. on Machine Learn. (ICML), pp. 2100–2111. Online.
    """
    def __init__(self, SSM: FilteringModel = None,
                 regularisation: float = 0.1,
                 decay_rate: float = 0.9,
                 min_update_size: float = 0.01,
                 max_iterations: int = 100,
                 transport_gradient_clip: float = 1.,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False
                 ) -> None:


        super().__init__(OptimalTransportResampler(regularisation, decay_rate, min_update_size, max_iterations, transport_gradient_clip), SSM, use_REINFORCE_for_proposal = use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)


class StopGradientDPF(ParticleFilter):
    """Differentiable particle filter with stop-gradient resampling.

    Parameters
    ----------
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    resampling_generator: torch.Generator
        Generator to track the resampling rng.
    multinomial: bool. Default False.
        If true then use multinomial resampling. Otherwise, use systematic resampling.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

    Notes
    -----
    See [1]_. If ``use_REINFORCE_for_proposal`` is ``True`` then this DPF is as described in [1]_ and the proposal model cannot be learned. Otherwise, it is as in [2]_.

    References
    ----------
    .. [1] Scibior A, Wood F (2021). “Differentiable particle filtering without modifying the forward pass.” arXiv:2106.10314

    .. [2] Cox B, P´erez-Vieites S, Zilberstein N, Sevilla M, Segarra S, Elvira V (2024). “End-to-End Learning of Gaussian Mixture Proposals Using Differentiable Particle Filters and Neural Networks.”
    In Int. Conf. Acoustics, Speech and Sig. Proc. (ICASSP), pp. 9701–9705.
    """
    def __init__(self, SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial:bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False
                 ) -> None:
        base_resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        resampler = StopGradientResampler(base_resampler)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)

class MarginalStopGradientDPF(MarginalParticleFilter):
    """Differentiable particle filter with marginal stop-gradient resampling.

    Parameters
    ----------
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    resampling_generator: torch.Generator
        Generator to track the resampling rng.
    multinomial: bool. Default False.
        If true then use multinomial resampling. Otherwise, use systematic resampling.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

    Notes
    -----
    See [1]_. If ``use_REINFORCE_for_proposal`` is ``True`` then this DPF is as described in [1]_ and the proposal model cannot be learned. Otherwise, if the model is bootstrap it is as in [2]_, if
    the SSM has a proposal then this is a non-published scheme.

    References
    ----------
    .. [1] Scibior A, Wood F (2021). “Differentiable particle filtering without modifying the forward pass.” arXiv:2106.10314

    .. [2] Brady JJ, et al. (2024). “Differentiable Interacting Multiple Model Particle Filtering.” arXiv preprint arXiv:2410.00620.
    """

    def __init__(self,
                 SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial = False,
                 *,
                 use_REINFORCE_for_proposal:bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False,
                 ) -> None:

        base_resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        resampler = StopGradientResampler(base_resampler)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)


class KernelDPF(ParticleFilter):
    """Differentiable particle filter with mixture kernel resampling.

    Parameters
    ----------
    SSM: FilteringModel
        A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
    Kernel: KernelMixture
        The Kernel mixture to resample from.
    use_REINFORCE_for_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
    use_REINFORCE_for_initial_proposal: bool. Default False.
        Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.


    Notes
    -----
    See [1]_. Currently only Gaussian Kernels are supported.

    References
    ----------
    .. [1] Younis A, Sudderth E (2023). “Differentiable and Stable Long-Range Tracking of Multiple Posterior Modes.” In Proc. Adv. Neural Inf. Process. Syst. (NeurIPS), volume 36. New Orleans, LA, USA.
    """


    def __init__(self,
                 SSM: FilteringModel = None,
                 kernel: KernelMixture = None,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False) -> None:

        if kernel is None:
            raise ValueError('Must specify a kernel mixture')

        super().__init__(KernelResampler(kernel), SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)


class MarginalDPF(MarginalParticleFilter):
    """Differentiable particle filter based on the marginal particle filter.

        Notes
        -----
        Analagous to the basic 'differentiable' particle filter, as described in R. Jonschkowski, D. Rastogi and O. Brock
        'Differentiable Particle Filters: End-to-End Learning with Algorithmic Priors' 2018 but based on the marginal particle filter.

        Parameters
        ----------
        SSM: FilteringModel
            A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
        resampling_generator: torch.Generator
            Generator to track the resampling rng.
        multinomial: bool. Default False.
            If true then use multinomial resampling. Otherwise, use systematic resampling.
        use_REINFORCE_for_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
        use_REINFORCE_for_initial_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

        See Also
        --------
        MarginalParticleFilter : The marginal particle filter base implementation.
        """

    def __init__(self,
                 SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial: bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False
                 ) -> None:
        resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)
        temp = copy(self.proposal)
        self.proposal = lambda prev_state, prev_weight, **data: temp(prev_state.detach(), prev_weight.detach(), **data)

class StraightThroughDPF(ParticleFilter):
    """Similar to the DPF but the gradient of the state is passed through resampling without modification. (T. Le et al. 'Auto-encoding sequential monte carlo' 2018,
        C. Maddison et al. ' Filtering variational objectives' 2018, and C. Naesseth et al. 'Variational sequential monte carlo' 2018.) Equivelant to soft resampling
        with a softness parameter of 1.

        Parameters
        ----------
        SSM: FilteringModel
            A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
        resampling_generator: torch.Generator
            Generator to track the resampling rng.
        multinomial: bool. Default False.
            If true then use multinomial resampling. Otherwise, use systematic resampling.
        use_REINFORCE_for_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
        use_REINFORCE_for_initial_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.
    """

    def __init__(self,
                 SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial:bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False) -> None:
        """


        Parameters
        ----------
        SSM: FilteringModel
            A FilteringModel that represents the SSM (and optionally a proposal model). See the documentation of FilteringModel for more complete information.
            If this parameter is not None then the values of initial_proposal and proposal are ignored.
        resampling_generator:
            The generator to track the resampling rng.
        """
        resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)


class MarginalStraightThroughDPF(MarginalParticleFilter):
    """The marginal particle filter with the straight through gradient estimator.


        Parameters
        ----------
        SSM: FilteringModel
            A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
        resampling_generator: torch.Generator
            Generator to track the resampling rng.
        multinomial: bool. Default False.
            If true then use multinomial resampling. Otherwise, use systematic resampling.
        use_REINFORCE_for_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
        use_REINFORCE_for_initial_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

        See Also
        --------
        MarginalParticleFilter : The marginal particle filter base implementation.
        StraightThroughDPF : The DPF with the straight through gradient estimator.
    """

    def __init__(self,
                 SSM: FilteringModel = None,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial: bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False) -> None:
        resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal)




class MarginalSoftDPF(MarginalParticleFilter):
    """Marginal particle filter with soft-resampling.

        Parameters
        ----------
        SSM: FilteringModel
            A ``FilteringModel`` that represents the SSM (and optionally a proposal model). See the documentation of ``FilteringModel`` for more complete information.
        resampling_generator: torch.Generator
            Generator to track the resampling rng.
        multinomial: bool. Default False.
            If true then use multinomial resampling. Otherwise, use systematic resampling.
        use_REINFORCE_for_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the particle proposal process.
        use_REINFORCE_for_initial_proposal: bool. Default False.
            Whether to use the REINFORCE estimator for the gradient due to the initial particle proposal process.

        See Also
        --------
        MarginalParticleFilter : The marginal particle filter base implementation.
        SoftDPF : The soft DPF
        """

    def __init__(self, SSM: FilteringModel = None,
                 softness: float = 0.7,
                 resampling_generator: torch.Generator = torch.default_generator,
                 multinomial: bool = False,
                 *,
                 use_REINFORCE_for_proposal: bool = False,
                 use_REINFORCE_for_initial_proposal: bool = False
                 ) -> None:
        base_resampler = MultinomialResampler(resampling_generator) if multinomial else SystematicResampler(resampling_generator)
        resampler = SoftResampler(softness, base_resampler, resampling_generator.device)
        super().__init__(resampler, SSM, use_REINFORCE_for_proposal=use_REINFORCE_for_proposal, use_REINFORCE_for_initial_proposal=use_REINFORCE_for_initial_proposal, optimise_for_bootstrap=False)


"""Python module that implements the Kalman filter using the Distributions submodule"""
import torch
from torch import Tensor
from .base import Module
from .distributions import LinearGaussian, MultivariateGaussian


class KalmanFilter(Module):
    """The standard Kalman filter for exact filtering of linear and Gaussian SSMs.

    Parameters
    ----------
    prior_model: distributions.MultivariateGaussian
        The prior model
    dynamic_model: distributions.LinearGaussian
        The dynamic model
    observation_model: distributions.MultivariateGaussian
        The observation model

    Notes
    -----
    See [1]_.

    References
    ----------
    .. [1] Kalman RE (1960). “A new approach to linear filtering and prediction problems.” Transactions of the ASME–Journal of Basic Engineering, 82(Series D), 35–45.
    """

    half_log_2pi = (1 / 2) * torch.log(torch.tensor(2 * torch.pi))

    def __init__(self, prior_model:MultivariateGaussian = None, dynamic_model:LinearGaussian=None, observation_model:LinearGaussian=None):
        super().__init__()
        self.prior_model = prior_model
        self.dynamic_model = dynamic_model
        self.observation_model = observation_model

    @staticmethod
    def _get_time_data(t: int, **data: dict, ) -> dict:
        time_dict = {k: v[t] for k, v in data.items() if k != 'series_metadata' and k != 'state' and v is not None}
        if data['time'] is not None and t > 0:
            time_dict['prev_time'] = data['time'][t - 1]
        if data['series_metadata'] is not None:
            time_dict['series_metadata'] = data['series_metadata']
        return time_dict

    @staticmethod
    def _Gaussian_log_density(mean, covariance, sample) -> Tensor:
        prefactor = -sample.size(-1) * KalmanFilter.half_log_2pi - torch.slogdet(covariance)[1]/2
        residuals = sample - mean
        exponent = (-1/2) * (residuals.unsqueeze(-2) @  torch.linalg.inv_ex(covariance)[0] @ residuals.unsqueeze(-1)).reshape(-1)
        return prefactor + exponent

    @staticmethod
    def _Bayes_update(prior_mean, prior_covariance, observation, observation_weight, observation_bias, observation_kernel_covariance):
        cov_weight = prior_covariance @ observation_weight.T
        inverse_term = torch.linalg.inv_ex(observation_weight @ cov_weight + observation_kernel_covariance)[0]
        weight_inverse_term_weight = observation_weight.T @ inverse_term @ observation_weight
        posterior_covariance = prior_covariance @ (torch.eye(prior_mean.size(-1), device = prior_mean.device) -  weight_inverse_term_weight @ prior_covariance)
        mean_term_1 = (torch.eye(prior_mean.size(-1), device = prior_mean.device) - prior_covariance @ weight_inverse_term_weight).unsqueeze(0) @ prior_mean.unsqueeze(-1)
        mean_term_2 = (cov_weight @ inverse_term).unsqueeze(0) @ (observation - observation_bias.unsqueeze(0)).unsqueeze(-1)
        return (mean_term_1 + mean_term_2).squeeze(-1), posterior_covariance

    @staticmethod
    def _propagate(prev_mean, prev_covariance, kernel_weight, kernel_bias, kernel_covariance):
        new_mean = (kernel_weight.unsqueeze(0) @ prev_mean.unsqueeze(-1)).squeeze(-1) + kernel_bias.unsqueeze(0)
        new_covariance = (kernel_weight @ prev_covariance @ kernel_weight.T) + kernel_covariance
        return new_mean, new_covariance


    def forward(self, time_extent: int, observation: Tensor):
        r"""Returns the filtering posterior distributions :math:`p(x_{t}\mid y_{0:t})` from Kalman filtering.


        Parameters
        ----------
        time_extent: int
            The maximum time-step to run to, including time 0, the filter will calculate {time_extent + 1} filtering distributions.
        observation: Tensor
            The observations of the hidden variable system.

        Returns
        -------
        posterior_means: Tensor
            The means of the Gaussian filtering posteriors.
        posterior_covariances: Tensor
            The covariances of the Gaussian filtering posteriors.
        likelihood_factors: Tensor
            The likelihood factors of the Gaussian filtering posteriors. Equal to :math:`p(y_{t}\mid y_{0:t-1})`.
        """
        prior_mean = self.prior_model.mean
        prior_covariance = self.prior_model.cholesky_covariance @ self.prior_model.cholesky_covariance.T

        dynamic_weight = self.dynamic_model.mean_fun.weight
        dynamic_bias = self.dynamic_model.mean_fun.bias
        dynamic_covariance = self.dynamic_model.dist.cholesky_covariance @ self.dynamic_model.dist.cholesky_covariance.T

        observation_weight = self.observation_model.mean_fun.weight
        observation_bias = self.observation_model.mean_fun.bias
        observation_kernel_covariance = self.observation_model.dist.cholesky_covariance @ self.observation_model.dist.cholesky_covariance.T

        posterior_means = torch.empty((time_extent+1, observation.size(1), prior_mean.size(-1)), device = observation.device, dtype=observation.dtype)
        posterior_covariances = torch.empty((time_extent+1, observation.size(1), prior_mean.size(-1), prior_mean.size(-1)), device = observation.device, dtype=observation.dtype)
        likelihood_factors = torch.empty((time_extent+1, observation.size(1)), device = observation.device, dtype=observation.dtype)
        observation_mean, observation_covariance = self._propagate(prior_mean.unsqueeze(0), prior_covariance, observation_weight, observation_bias, observation_kernel_covariance)
        posterior_means[0], posterior_covariances[0] = self._Bayes_update(prior_mean.unsqueeze(0), prior_covariance, observation[0], observation_weight, observation_bias, observation_kernel_covariance)
        likelihood_factors[0] = self._Gaussian_log_density(observation_mean, observation_covariance, observation[0])

        for t in range(time_extent):
            predictive_mean, predictive_covariance = self._propagate(posterior_means[t], posterior_covariances[t], dynamic_weight, dynamic_bias, dynamic_covariance)
            observation_mean, observation_covariance = self._propagate(predictive_mean, predictive_covariance, observation_weight, observation_bias, observation_kernel_covariance)
            likelihood_factors[t+1] = self._Gaussian_log_density(observation_mean, observation_covariance, observation[t+1])
            posterior_means[t+1], posterior_covariances[t+1] = self._Bayes_update(predictive_mean, predictive_covariance, observation[t+1], observation_weight, observation_bias, observation_kernel_covariance)

        return posterior_means, posterior_covariances, likelihood_factors